/** @file
 * This file defines a rendezvous extension for the Host Identity Protocol
 * (HIP).
 * 
 * The rendezvous extension extends HIP and the HIP registration extension for
 * initiating communication between HIP nodes via HIP rendezvous servers. The
 * rendezvous server (RVS) serves as an initial contact point ("rendezvous
 * point") for its clients.  The clients of an RVS are HIP nodes that use the
 * HIP Registration Protocol
 * [<a href="http://www.ietf.org/internet-drafts/draft-ietf-hip-registration-02.txt">
 * draft-ietf-hip-registration-02</a>] to register their HIT->IP address mappings
 * with the RVS. After this registration, other HIP nodes can initiate a base
 * exchange using the IP address of the RVS instead of the current IP address
 * of the node they attempt to contact.
 * 
 * A rendezvous server stores the HIT->IP address mappings of its clients into
 * a hashtable as a rendezvous association data structure. A client can have
 * a maximum number of @c HIP_RVA_MAX_IPS IP addresses mapped to a single HIT,
 * and a RVS can have a maximum of @c HIP_RVA_SIZE clients.
 * 
 * Version 1.1 added function comments, removed many useless functions and
 * renewed the hip_rvs_relay_i1() -function.
 * 
 * @author  (version 1.0) Kristian Slavov
 * @author  (version 1.1) Lauri Silvennoinen
 * @version 1.1
 * @date    27.10.2006
 * @note    Related draft:
 *          <a href="http://tools.ietf.org/wg/hip/draft-ietf-hip-rvs/draft-ietf-hip-rvs-05.txt">
 *          draft-ietf-hip-rvs-05</a>
 * @note    Distributed under <a href="http://www.gnu.org/licenses/gpl.txt">GNU/GPL</a>.
 * @note    Version 1.0 was document scarcely and the comments regarding
 *          version 1.0 that have been added afterwards may be inaccurate.
 */ 
#include "rvs.h"

/** A hashtable for storing rendezvous associations. */
HIP_HASHTABLE *rva_table;
/** A linked list head used inside @c rva_table hashtable. */
static hip_list_t rvadb[HIP_RVA_SIZE];

/**
 * Allocates and initializes a rendezvous association.
 *
 * @param gfpmask a mask for HIP_MALLOC() that is used to allocate the memory.
 * @return        a pointer to a newly allocated and initialized rendezvous 
 *                association structure or NULL if failed to allocate memory.
 */
hip_rva_t *hip_rvs_allocate(int gfpmask)
{
	hip_rva_t *res;

	if((res = HIP_MALLOC(sizeof(*res), gfpmask)) == NULL) {
		HIP_ERROR("Error allocating memory for rendezvous association.\n");
		return NULL;
	}

	atomic_set(&res->refcnt, 0);
	HIP_LOCK_INIT(res);
	res->lifetime = 0; // HIP_DEFAULT_RVA_LIFETIME
	memset(res->ip_addrs, 0, HIP_RVA_MAX_IPS*sizeof(struct in6_addr));

	return res;
}

/**
 * Creates a rendezvous association from a host association.
 * 
 * Allocates memory for a new rendezvous association and copies information
 * from the parameter host association into it.
 * 
 * @param  ha      a pointer to a host association from where from to copy.
 * @param send_pkt a function pointer to a function to be used for relaying I1
 *                 packet.
 * @return         a pointer to a newly allocated rendezvous association or
 *                 NULL if failed to allocate memory.
 */
hip_rva_t *hip_rvs_ha2rva(hip_ha_t *ha, hip_xmit_func_t send_pkt)
{
	hip_rva_t *rva = NULL;
	struct hip_peer_addr_list_item *entry;
	int ipcnt = 0;
	struct hip_spi_out_item *spi_entry;
	hip_list_t *spi_out, *spi_tmp;
	hip_list_t *item, *tmp;
	int i, ii;

	_HIP_DEBUG("hip_rvs_ha2rva() invoked.\n");
	_HIP_DEBUG("ha->peer_udp_port:%d.\n", ha->peer_udp_port);

	if((rva = hip_rvs_allocate(GFP_KERNEL)) == NULL) {
		HIP_ERROR("Error allocating memory for rendezvous association.\n");
		return NULL;
	}
	
	/* Incremented the refrerence count of the new rendezvous association. */
	hip_hold_rva(rva);
	
	/* Lock the host association while copying values from it. */
	HIP_LOCK_HA(ha);

	/* Copy the client udp port. */
	if(ha->peer_udp_port != 0) {
		rva->client_udp_port = ha->peer_udp_port;
	}
	else {
		rva->client_udp_port = 0;
	}
	
	/* Copy peer hit as the client hit. */
	ipv6_addr_copy(&rva->hit, &ha->hit_peer);
	
	/* Copy HMACs. */
	memcpy(&rva->hmac_our, &ha->hip_hmac_in, sizeof(rva->hmac_our));
 	memcpy(&rva->hmac_peer, &ha->hip_hmac_out, sizeof(rva->hmac_peer));
	
	/* Set xmit-function. */
	rva->send_pkt = send_pkt;

	/* If the host association has a preferred address, copy it as the
	   first IP address of the rendezvous association. */
	if (!ipv6_addr_any(&ha->preferred_address)) {
		HIP_DEBUG("Copying base exchange address.\n");
		ipv6_addr_copy(&rva->ip_addrs[ipcnt], &ha->preferred_address);
		ipcnt++;
		if (ipcnt >= HIP_RVA_MAX_IPS)
			goto out;
	}

	/* Copy rest of the IP addresses. */
	list_for_each_safe(spi_out, spi_tmp, ha->spis_out, i)
	{
		spi_entry = list_entry(spi_out);
		list_for_each_safe(item, tmp, spi_entry->peer_addr_list, ii)
		{
			entry = list_entry(item);
			if (entry->address_state != PEER_ADDR_STATE_ACTIVE)
				continue;

			ipv6_addr_copy(&rva->ip_addrs[ipcnt], &entry->address);
			ipcnt++;
			if (ipcnt >= HIP_RVA_MAX_IPS)
				break;
		}
		if (ipcnt >= HIP_RVA_MAX_IPS)
			break;
	}
 out:
	HIP_UNLOCK_HA(ha);
	return rva;
}

/**
 * Gets a rendezvous association from the rendezvous association hashtable.
 *
 * Gets a rendezvous association matching the argument @c hit. If
 * a rendezvous association is found, it is automatically holded (refcnt
 * incremented).
 *
 * @param hit the HIT of the rendezvous association to get.
 * @return    a pointer to a matching rendezvous association or NULL if
 *            a matching rendezvous association was not found.
 */
hip_rva_t *hip_rvs_get(struct in6_addr *hit)
{
 	return (hip_rva_t*)hip_ht_find(rva_table, hit);
}

/**
 * Gets a valid rendezvous association from the rendezvous association
 * hashtable.
 *
 * Gets a valid rendezvous association matching the argument hit. Finds
 * a rendezvous association matching the argument hit and whose state is
 * @c HIP_RVASTATE_VALID. If a valid rendezvous association is found,
 * it is automatically holded (refcnt incremented).
 * 
 * @param hit the HIT of the rendezvous association to get.
 * @return a pointer to a matching valid rendezvous association or NULL if
 *         a matching valid rendezvous association was not found.
 */
hip_rva_t *hip_rvs_get_valid(struct in6_addr *hit)
{
	hip_rva_t *rva;

 	rva = (hip_rva_t*)hip_ht_find(rva_table, hit);
 	if (rva) {
 		if ((rva->rvastate & HIP_RVASTATE_VALID) == 0) {
			HIP_ERROR("A matching rendezvous association was found, "\
				  "but the state is not valid.\n");
 			hip_put_rva(rva);
 			rva = NULL;
 		}
 	}
	return rva;
}

/**
 * Inserts or updates a rendezvous association IP address.
 * 
 * A rendezvous server client can register more than one of its IP addresses to
 * a rendezvous server. In this case the rendezvous association has a maximum
 * number of @c HIP_RVA_MAX_IPS IP addresses mapped to a single HIT.
 * This function inserts one IP address to a rendezvous association. The index
 * of the IP address to insert must be smaller than @c HIP_RVA_MAX_IPS. The
 * existing IP at @c index is overwritten.
 *
 * @param rva   the rendezvous association whose IP address is to be modified. 
 * @param ip    the IP address to insert.
 * @param index the index of the IP address to be modified.
 */
void hip_rvs_put_ip(hip_rva_t *rva, struct in6_addr *ip, unsigned int index)
{
	HIP_ASSERT(rva);
 	HIP_ASSERT(index < HIP_RVA_MAX_IPS);

 	HIP_LOCK_HA(rva);
 	ipv6_addr_copy(&rva->ip_addrs[index], ip);
 	HIP_UNLOCK_HA(rva);
}

/**
 * Gets an IP address from a rendezvous association.
 * 
 * Gets an IP address at @c index from a rendezvous association. Destination
 * buffer @c dst must be allocated before calling this function and @c index
 * must be smaller than @c HIP_RVA_MAX_IPS.
 *
 * @param rva   the rendezvous association from where to get the IP address.
 * @param dst   a pointer to a buffer where to put the IP address.
 * @param index the index of the IP address to get.
 */
void hip_rvs_get_ip(hip_rva_t *rva, struct in6_addr *dst, unsigned int index)
{
	HIP_ASSERT(rva);
	HIP_ASSERT(dst);
	HIP_ASSERT(index < HIP_RVA_MAX_IPS);

	HIP_LOCK_HA(rva);
	ipv6_addr_copy(dst, &rva->ip_addrs[index]);
	HIP_UNLOCK_HA(rva);
}

/**
 * Inserts a new rendezvous association into the rendezvous association
 * hashtable.
 * 
 * Inserts a parameter rendezvous association @c rva into the rendezvous
 * association hashtable. The rendezvous association @c rva is automatically
 * holded (refcnt incremented).
 *
 * @param rva the rendezvous association to be added into the hashtable.
 * @return    zero on success, or negative error value on error.
 */
int hip_rvs_put_rva(hip_rva_t *rva)
{
	int err;
	HIP_DEBUG_HIT("hip_rvs_put_rva(): Inserting rendezvous association "\
		      "with hit", &rva->hit);
	
	/* If assertation holds, then we don't need locking */
	HIP_ASSERT(atomic_read(&rva->refcnt) <= 1); 

	HIP_IFEL(ipv6_addr_any(&rva->hit), -EINVAL,
		 "Trying to insert rva entry with NULL hit.\n");
	HIP_IFEL(hip_ht_find(rva_table, &rva->hit), -EEXIST,
		 "Duplicate rva entry. Not adding hit to rva table.\n");
	
	err = hip_ht_add(rva_table, rva);
	if (err) {
		HIP_ERROR("Failed to add rva hash table entry.\n");
	} else {
		rva->rvastate |= HIP_RVASTATE_VALID;
	}
 out_err:
	return err;
}

/**
 * Initializes the rendezvous association hashtable.
 *
 * Initializes the rendezvous association hashtable @c rva_table by
 * <ul>
 * <li>setting the memory area all zeros,
 * <li>setting the hashsize and offset,
 * <li>setting the needed function pointers,
 * <li>naming the table as <code>RVA TABLE</code>.
 * </ul>
 */ 
void hip_rvs_init_rvadb()
{
	rva_table = hip_ht_init(hip_hash_hit, hip_match_hit);
	if (!rva_table) HIP_ERROR("hip_rvs_init_rvadb() error!\n");
}

/**
 * Uninitializes the rendezvous association hashtable.
 * 
 * Uninitializes the rendezvous association hashtable @c rva_table by calling
 * hip_ht_uninit() for @c rva_table. Memory allocated for each rendezvous
 * association is freed.
 */ 
void hip_rvs_uninit_rvadb()
{
	hip_ht_uninit(rva_table);
}

/**
 * Frees the memory allocated for a rendezvous association.
 *
 * Frees the memory allocated for the parameter rendezvous association.
 *
 * @param rva the rendezvous association whose memory to free.
 * @note      this function should be called only after the last reference to
 *            the parameter @c rva is deleted. 
 */ 
void hip_rvs_free_rva(hip_rva_t *rva)
{
	HIP_FREE(rva);
}

/**
 * Removes a rendezvous association from the rendezvous association hashtable.
 *
 * Removes the parameter rendezvous association from the rendezvous
 * association hashtable.
 *
 * @param rva the rendezvous association to remove.
 */ 
void hip_rvs_remove(hip_rva_t *rva)
{
	HIP_ASSERT(rva);

	HIP_LOCK_HA(rva);
	if (!(rva->rvastate & HIP_RVASTATE_VALID)) {
		HIP_DEBUG("RVA not in rva hashtable or state corrupted\n");
		return;
	}

	hip_ht_delete(rva_table, rva);
	HIP_UNLOCK_HA(rva);

	/* the refcnt should now be at least 1, since we must have at least two
	   references when calling this function: One in the hashtable, one in
	   the calling function. After the delete operation we should have one
	   less, but still over 0 and thus the HIP_UNLOCK_HA should not cause
	   problems (accessing deleted memory). */
}

/**
 * Relays an incoming I1 packet.
 *
 * This function relays an incoming I1 packet to the next node on path
 * to receiver and inserts a @c FROM parameter encapsulating the source IP
 * address. In case there is a NAT between the sender (the initiator or previous
 * RVS) of the I1 packet, a @c FROM_NAT parameter is inserted instead of a
 * @c FROM parameter. Next node on path is typically the responder, but if the
 * message is to travel multiple rendezvous servers en route to responder, next
 * node can also be another rendezvous server. In this case the @c FROM
 * (@c FROM_NAT) parameter is appended after the existing ones. Thus current RVS
 * appends the address of previous RVS and the final RVS (n) in the RVS chain
 * sends @c FROM:I, @c FROM:RVS1, ... , <code>FROM:RVS(n-1)</code>. If initiator
 * is located behind a NAT, the first @c FROM parameter is replaced with a
 * @c FROM_NAT parameter.
 * 
 * @param i1       a pointer to the I1 HIP packet common header with source and
 *                 destination HITs.
 * @param i1_saddr a pointer to the source address from where the I1 packet was
 *                 received.
 * @param i1_daddr a pointer to the destination address where the I1 packet was
 *                 sent to (own address).
 * @param rva      a pointer to a rendezvous association matching the HIT of
 *                 next hop.
 * @param i1_info  a pointer to the source and destination ports (when NAT is
 *                 in use).
 * @return         zero on success, or negative error value on error.
 * @note           This code has not been tested thoroughly with multiple RVSes.
 */
int hip_rvs_relay_i1(const struct hip_common *i1,
		     const struct in6_addr *i1_saddr,
		     const struct in6_addr *i1_daddr, hip_rva_t *rva,
		     const hip_portpair_t *i1_info)
{
	struct hip_common *i1_to_be_relayed = NULL;
	struct hip_tlv_common *current_param = NULL;
	int err = 0, from_added = 0;
	struct in6_addr final_dst;
	hip_tlv_type_t param_type = 0;
	/* A function pointer to either hip_build_param_from() or
	   hip_build_param_from_nat(). */
	int (*builder_function) (struct hip_common *msg,
				 const struct in6_addr *addr,
				 const in_port_t port);

	_HIP_DEBUG("hip_rvs_relay_i1() invoked.\n");
	HIP_DEBUG_IN6ADDR("hip_rvs_relay_i1():  I1 source address", i1_saddr);
	HIP_DEBUG_IN6ADDR("hip_rvs_relay_i1():  I1 destination address", i1_daddr);
	HIP_DEBUG_HIT("hip_rvs_relay_i1(): Rendezvous association hit", &rva->hit);
	HIP_DEBUG("Rendezvous association port: %d.\n", rva->client_udp_port);
	HIP_DEBUG("I1 source port: %u, destination port: %u\n",
		  i1_info->src_port, i1_info->dst_port);
		
	/* If the incoming I1 packet was destined to port 50500, we know that
	   there is a NAT between (I->NAT->RVS->R). */
	if(i1_info->dst_port == HIP_NAT_UDP_PORT) {
		builder_function = hip_build_param_from_nat;
		param_type = HIP_PARAM_FROM_NAT;
	}
	else {
		builder_function = hip_build_param_from;
		param_type = HIP_PARAM_FROM;
	}

	/* Get the destination IP address which the client has registered from
	   the rendezvous association. */
	/** @todo How to decide which IP address of rva->ip_addrs the to use? */
	hip_rvs_get_ip(rva, &final_dst, 0);

	HIP_IFEL(!(i1_to_be_relayed = hip_msg_alloc()), -ENOMEM,
		 "No memory to copy original I1\n");	

	/* I1 packet forwarding is achieved by rewriting the source and
	   destination IP addresses. */
	hip_build_network_hdr(i1_to_be_relayed, HIP_I1, 0,
			      &(i1->hits), &(i1->hitr));

	/* Adding FROM (FROM_NAT) parameter. Loop through all the parameters in
	   the received I1 packet, and insert a new FROM (FROM_NAT) parameter
	   after the last found FROM (FROM_NAT) parameter. Notice that in most
	   cases the incoming I1 has no paramaters at all, and this "while" loop
	   is skipped. Multiple rvses en route to responder is one (and only?)
	   case when the incoming I1 packet has parameters. */
	while ((current_param = hip_get_next_param(i1, current_param)) != NULL){
		
		HIP_DEBUG("Found parameter in I1.\n");
		/* Copy while type is smaller than or equal to FROM (FROM_NAT)
		   or a new FROM (FROM_NAT) has already been added. */
		if (from_added || hip_get_param_type(current_param) <= param_type)
		{
			HIP_DEBUG("Copying existing parameter to I1 packet "\
				  "to be relayed.\n");
			hip_build_param(i1_to_be_relayed,current_param);
			continue;
		}
		/* Parameter under inspection has greater type than FROM
		   (FROM_NAT) parameter: insert a new FROM (FROM_NAT) parameter
		   between the last found FROM (FROM_NAT) parameter and
		   "current_param". */
		else
		{
			HIP_DEBUG("Created new %s and copied "\
				  "current parameter to relayed I1.\n",
				  hip_param_type_name(param_type));
			builder_function(i1_to_be_relayed, i1_saddr,
					 i1_info->src_port);
			hip_build_param(i1_to_be_relayed, current_param);
			from_added = 1;
		}
	}

	/* If the incoming I1 had no parameters after the existing FROM (FROM_NAT)
	   parameters, new FROM (FROM_NAT) parameter is not added until here. */
	if (!from_added)
	{
		HIP_DEBUG("No parameters found, adding a new %s.\n",
			  hip_param_type_name(param_type));
		builder_function(i1_to_be_relayed, i1_saddr, i1_info->src_port);
	}

	/* Zero message HIP checksum. */
	hip_zero_msg_checksum(i1_to_be_relayed);

	/* Adding RVS_HMAC parameter as the last parameter of the relayed
	   packet. Notice, that this presumes that there are no parameters
	   whose type value is greater than RVS_HMAC in the incoming I1
	   packet. */
	HIP_DEBUG("Adding a new RVS_HMAC parameter as the last parameter.\n");
	HIP_IFEL(hip_build_param_rvs_hmac_contents(i1_to_be_relayed,
						   &rva->hmac_our), -1,
		 "Building of RVS_HMAC failed.\n");
	
	/* If the client is behind NAT the I1 packet is relayed on UDP. If
	   there is no NAT the packet is relayed on raw HIP. We don't have to
	   take care of which send-function to use, as the rva->send_pkt was
	   initiated with correct value in hip_rvs_ha2rva() when the rva was
	   created. Note that we use NULL as source IP address instead of
	   i1_daddr. A source address is selected in the corresponding
	   send-function. */
	HIP_IFEL(rva->send_pkt(NULL, &final_dst, HIP_NAT_UDP_PORT,
			       rva->client_udp_port, i1_to_be_relayed, NULL, 0),
		 -ECOMM, "Relaying I1 failed.\n");
	
	HIP_DEBUG_HIT("hip_rvs_relay_i1(): Relayed I1 to", &final_dst);

 out_err:
	if(i1_to_be_relayed != NULL)
	{
		HIP_FREE(i1_to_be_relayed);
	}
	return err;
}


/**
 * Sends a UDP encapsulated NOTIFY packet to the initiator.
 * 
 * Sends a UDP encapsulated NOTIFY packet with a NOTIFICATION parameter of type
 * RVS_NAT to the initiator located at @c i1_saddr. NOTIFICATION parameter has
 * responder's HIT, IP and port number.
 *
 * @param i1       a pointer to the I1 HIP packet common header with source and
 *                 destination HITs.
 * @param i1_saddr a pointer to the source address from where the I1 packet was
 *                 received.
 * @param rva      a pointer to a rendezvous association matching the HIT of
 *                 next hop.
 * @param i1_info  a pointer to the source and destination ports (when NAT is
 *                 in use).
 * @return         zero on success, or negative error value on error.
 */ 
int hip_rvs_reply_with_notify(const struct hip_common *i1,
			      struct in6_addr *i1_saddr,
			      hip_rva_t *rva,
			      const hip_portpair_t *i1_info,
			      uint16_t param_type)
{
	int err = 0; 
	struct hip_common *notify = NULL;
	struct in6_addr responder_ip;
	void *data = NULL;
	size_t data_len = 2 * sizeof(struct in6_addr) + sizeof(in_port_t);
	_HIP_DEBUG("hip_rvs_reply_with_notify() invoked.\n");

	HIP_IFEL((notify = hip_msg_alloc()) == NULL, -ENOMEM,
		 "No memory to create a NOTIFY packet.\n");
	
	HIP_IFEL((data = HIP_MALLOC(data_len, 0)) == NULL,
		 -ENOMEM, "No memory for notification data.\n");
	
	/* Get the destination IP address which the client has registered from
	   the rendezvous association. */
	/** @todo How to decide which IP address of rva->ip_addrs the to use? */
	hip_rvs_get_ip(rva, &responder_ip, 0);
	hip_build_network_hdr(notify, HIP_NOTIFY, 0, &(i1->hitr), &(i1->hits));
	
	/* Create NOTIFICATION data. */
	ipv6_addr_copy(data, &(i1->hitr));
	data += sizeof((i1->hitr));
	ipv6_addr_copy(data, &responder_ip);
	data += sizeof(responder_ip);
	memcpy(data, &(rva->client_udp_port), sizeof(rva->client_udp_port));
	data -= sizeof((i1->hitr)) + sizeof(responder_ip);
	
	hip_build_param_notification(notify, param_type, data, data_len); 
	
	HIP_IFEL(hip_send_udp(NULL, i1_saddr, HIP_NAT_UDP_PORT,
			      i1_info->src_port, notify,
			      NULL, 0),
		 -ECOMM, "Sending NOTIFY packet on UDP failed.\n");
	
 out_err:
	if (notify != NULL) {
		HIP_FREE(notify);
	}
	if (data != NULL) {
		HIP_FREE(data);
	}
	return err;
}

/**
 * Set a rendezvous server request flag for a host association.
 *
 * Set a rendezvous server request flag for a host association matching the
 * source/destination -hitpair. Calling this function indicates that the current
 * machine is requesting a rendezvous service for a host association entry. This
 * function is called when a host starts registering procedure to rendezvous
 * service by sending a @b I1 packet to the rendezvous server.
 *
 * @param  src_hit the source hit of a host association.
 * @param  dst_hit the destination hit of a host association.
 * @return zero on success, or negative error value if a matching entry is
 *         not found.
 */ 
int hip_rvs_set_request_flag(hip_hit_t *src_hit,
			     hip_hit_t *dst_hit)
{
	int err = 0;
	hip_ha_t *entry;
	
	HIP_IFEL(!(entry = hip_hadb_find_byhits(src_hit, dst_hit)),
		 -1, "Could not set RVS request bit\n");

	entry->local_controls |= HIP_PSEUDO_CONTROL_REQ_RVS;
	hip_put_ha(entry);

 out_err:
	return err;
}
